{% extends "base.html" %}

{% block content %}

    <h1 id=header>Your experiments</h1>

    {{ experimenttable() }}

    <script>
        reload(() => data_ajax('get_user_experiments'), fillExperimentTable)
           
 
        function deleteAllExperiments() {
            if (confirm("This will delete ALL of your experiments and CANNOT be undone. Proceed?")) {
                    document.getElementById("deleteButton").innerHTML="<i class='fa fa-cog fa-spin'></i> Deleting..."
                    window.location.href = /delete_all/
            }
        }

    </script>
{% endblock %}
