{% extends "base.html" %}

{% block content %}

        <h1 id=header>All experiments</h1>

        {{ experimenttable() }}
        
        <script>
            reload( () => data_ajax('get_all_experiments'), 
                                       fillExperimentTable )

        </script>



{% endblock %}
